function [sigi]=LG_interp_WLDS(sig,ind_ini,ind_end,lmc,lw)
% function [xi]=LG_interp_WLDS(x,ind_ini,ind_end)
% This uses the Wavelet toolbox for signal analysis and synthesis (functions wavedec.m and waverec.m)
% INPUTS
% sig: input signal
% ind_ini: index of the first sample of the passage to be reconstructed (interpolated)
% ind_end: index of the last sample of the passage to be reconstructed (interpolated)
% lmc: number of uncorrupted signal samples preceding and succeeding the
% gap that will bw used for estimation of the model parameters
% lw: length in samples of the Daubechies prototype filter (8 for db8, 16 for db16 and so on) 
% OUTPUT
% sigi: interpolated signal

% Important observations:

% 1) ind_end > ind_ini 

% 2) Let lg = ind_end - ind_ini: 
% a) lg should be long (about 200 times the order of the AR model (50) used in the interpolation)
% b) lmc is recommended to be 1.5 x lg to 3 x lg samples 

% 3) All other processing parameters are predifined inside this function. 
% More information on the preseting of these parameters can be found from the paper
% P. A. A. Esquef and L. W. P. Biscainho, "An Efficient Model-Based Method for Reconstruction
% of Audio Signals across  Long Gaps", IEEE Trans. Speech and Language 
% Processing, vol. 14, no. 4, pp. 1391-1400, 2006. 
% Link:
% http://ieeexplore.ieee.org/iel5/9346/29688/01350530.pdf?arnumber=1350530 

% 4) There are no verifications of inconsistencies related to the use of this function, e.g.,
% signal too short, ind_end < ind_ini, or insufficient samples before and
% after the part to be reconstructed.

% PROCESSING PARAMETERS
% parameters for full band interpolation 
pfb=50;       % order of the AR model for the full-band interpolation (original samplerate)

% Parameters for sub-band intepolation 
bd=6; % number of sub-bands 
pv=[50 50 10 10 10 10];     % model orders in each subband (low -> high freq)

% Example: in a 6-subband decomposition:
% a) The signal of the channel 6 (highest frequency), model-based signal
% interpolation will be carried out with AR models of order 10 (pv(6)).
% b) The signal of the channel 1 (lowest frequency), model-based signal
% interpolation will be carried out with AR models of order 50 (pv(1)).

% Parameters for dual processing 
psb=2;   % number of low-frequency sub-bands to be processed. The highest
% (high-frequency subband signals) are left intact (only processed by the
% fullband interpolation).

% Defines type of sub-band decomposition (prototype filter)
type='db'; % Daubechies wavelets 
wtype=[type,num2str(lw)];

sigi=sig;  % stores in sigi the original signal (corrupted)

lg=ind_end+1-ind_ini;  % length of the gap in samples 
x=sig(ind_ini-lmc-1+(1:(2*lmc+lg)));  % portion of sig that is taken to be processed
[x]=gap_interpolation(x,pfb,lg,lmc+1,lmc); % gap interpolation in full-band  
% Begining of the subband interpolation:
N=bd-1; % number of subbands minus 1
NS=bd;
[C,L] = wavedec(x,N,wtype);  % C contains the subband signals and L indicators of their location in C
%D=zeros(L(end-1),NS);   
CL=[0; cumsum(L)];  % auxiliary vector used to indicate the beginning of each subband signal in C
% Example:
%cn=1;  % channel number (by convention 1 is the lowsest frequency one)
%rg=1:L(cn); % channel range in C
%cs=C(CL(cn)+(rg));  % channel signal in channel cn

% With reference to x the indices of the beginning and end of the gap are
% located at
ind_ini_o=lmc+1;  
ind_end_o=ind_ini_o+lg;
for cn=1:psb,  % for the first psb channels, do process the subband signals
    rg=1:L(cn); % channel range in C
    cs=C(CL(cn)+(rg));  % channel signal in channel cn
    [efl]=calculate_filter_length(lw,cn,NS); % calculates equivalent filter length in channel cn
    exp_dec=min(N,NS-cn+1); % calculate the exponent of the nominal length reduction due to downsalpling  
    ind_ini_ss=floor(ind_ini_o/2^exp_dec)+efl;  % beginning of the gap in the cn channel
    ind_end_ss=ceil(ind_end_o/2^exp_dec)+efl;  % end of the gao in the cn channel
    %ind_ini_ss=floor(ind_ini_o/2^exp_dec)+1;  % beginning of the gap in the cn channel
    %ind_end_ss=ceil(ind_end_o/2^exp_dec)+2*efl;  % end of the gao in the cn channel
    lmc_ss=floor(lmc/2^exp_dec);  % number of samples (after downsampling) before or after the gap used for model estimation 
    lgss=ind_end_ss+1-ind_ini_ss; % % number of samples (after downsampling) of the gap
    cs=gap_interpolation(cs,pv(cn),lgss,ind_ini_ss,lmc_ss); % interpolation at cn subband 
    C(CL(cn)+(rg))=cs; % writes back processed signal in C
    clear rg cs efl exp_dec ind_ini_ss ind_end_ss lmc_ss lgss 
end
xsint = waverec(C,L,wtype); % synthesizes a new signal from the PROCESSED subands signals 
% End of subband interpolation
sigi(ind_ini-lmc-1+(1:(2*lmc+lg)))=xsint; % replaces the processed passage in sigi



%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliary functions are defined below 
%%%%%%%%%%%%%%%%%%%%%%

function [xi]=gap_interpolation(x,p,lg,gi,lmc)
% function [xi]=gap_interpolation(x,p,lg,gi,lmc)
% x: input signal with a gap
% p: order of the AR model used in the interpolation 
% lg: length of the gap
% gi: index of the beginning of the gap
% lmc: number of samples used to estimate the AR models

% Uncomment the lines below to check if the gap location is correct
% plot(x); hold on
% axis tight
% aa=axis;
% plot([gi gi],[aa(3) aa(4)],'k--');
% plot([gi+lg gi+lg],[aa(3) aa(4)],'k--');
% pause; close

w=gw_win(lg,3);  % generates a window for the crossfading procedure
x=x(:);
s_ini=gi-lmc;  % index of the beginning of the signal used for model estimation
xf=x(s_ini-1+(1:lmc));            % segment used to compute the forward prediction
% For the AR model estimation it is necessary that xf be longer than model order p
pM=min(length(xf)-1,p);   % reduces model order to the maximum allowed
ab=arburg(xf,pM);                  % AR model estimation via Burg's method
Z=filtic(1,ab,flipud(xf(:)));     % obtains initial conditions of the filter
ex=filter(ab,1,xf); ex(1:p)=0;    % obtains the model error (or excitation signal) and sets the transient to zero 

tr_ex=flipud(ex(:));              % the excitation signal for the extrapolation model as the time-reversed version of ex 
% tr_ex should be at least lg-samples long
if length(tr_ex)<lg,
    aux=tr_ex;
    tr_ex=zeros(lg,1);
    tr_ex(1:length(aux))=aux;
end
fes=filter(1,ab,tr_ex,Z);         % creates the forward extrapolated signal 

wfes=fes(1:lg).*w;                % fades-out the forward extrapolated signal 
clear ab ex tr_ex Z aux pM
% Backward extrapolation 

i_xb1=gi+lg;                      % index of the beginning of the segment after the gap 
i_xb2=min(length(x),gi+lg+lmc);   % index of the end of the segment after the gap used for model estimation
xb=x(i_xb1:i_xb2);                % segment after the gap used for model estimation
xb=flipud(xb(:));                 % applies time-reversal to use the same procedure as in the forward extrapolation
% For the AR model estimation it is necessary that xb be longer than model order p
pM=min(length(xb)-1,p);   % reduces model order to the maximum allowed
ab=arburg(xb,pM);                  % AR model estimation via Burg's method  
Z=filtic(1,ab,flipud(xb(:)));     % obtains initial conditions of the filter
ex=filter(ab,1,xb); ex(1:p)=0;    % obtains the model error (or excitation signal) and sets the transient to zero 
tr_ex=flipud(ex(:));              % the excitation signal for the extrapolation model as the time-reversed version of ex 
% tr_ex should be at least lg-samples long
if length(tr_ex)<lg,
    aux=tr_ex;
    tr_ex=zeros(lg,1);
    tr_ex(1:length(aux))=aux;
end
bes=filter(1,ab,tr_ex,Z);         % creates the backward extrapolated signal 
bes=bes(:);                       % makes bes a column vector
wbes=flipud(bes(1:lg).*w);        % fades-in the borward extrapolated signal 

es=wfes+wbes;                     % sums the windowed forward and backward extrapolated signals 

xi=x;
xi(gi-1+(1:lg))=es;     % reconstructed signal

%%%%%%%%%%%%%%%%%%%%%%%%

function [w]=gw_win(n,a)

% this function implements the general weighting function

% n: is the length of the window
% a: is the damping parameter: a=1 implies linear slope decrease whereas 
% a=infinity implies a step function

ns=1;
nm1=n-ns;
u=zeros(1,n);
w=u;
for k=1:n,
	u(k)=(k-ns)/nm1;
        if u(k)<=0.5,
		w(k)=1-0.5*(2*u(k)).^a;
        elseif u(k)>0.5,
                w(k)=0.5*(2-2*u(k)).^a;
        end
end

w=w(:);

function [efl]=calculate_filter_length(ol,cn,NS)
% ol: original filter length (power of 2)
% cn: channel number: convention 1 is lowest freq 
% NS: number of subbands
% efl: equivalent filter length in channel cn 

cn=NS-cn-1;
if cn==-1,
    efl=4;
else
    v=0:cn; v=.5.^v;
    a=1/v(end);
    efl=.5*ol*sum(v)-(a-1)/a;
    efl=ceil(efl);
end
if cn==NS-2,
    efl=efl+round(.6875*ol);
end

